package Presenter.Algorithms;

import Model.Interval;
import Model.Line;
import Model.Point;
import Presenter.Presenter;

import java.util.*;

/**
 * Implementierung verschiedener Algorithmen zur Berechnung von Ausgleichsgeraden.
 *
 * @Author: Armin Wolf
 * @Email: a_wolf28@uni-muenster.de
 * @Date: 28.05.2017.
 */
public class TheilSenEstimator extends Observable implements Algorithm {

    private final Double POSITIV_INF = 99999.0;
    private final Double NEGATIV_INF = -99999.0;


    private Presenter presenter;
    private ArrayList<Line> setOfLines;
    private ArrayList<Point> setOfIntersections;
    private ArrayList<Point> intervalIntersections;
    private ArrayList<Double> sampledIntersections;


    //wird benötigt um den y Achsenabschnitt zu Berechnen
    private ArrayList<Double> yCoordinates;
    private ArrayList<Double> xCoordinates;


    //Hilfsvariablen (siehe original Paper)
    private Double j;
    private Integer jA;
    private Integer jB;
    private Double r;
    private Integer n;
    private Double N;
    private Integer k;

    //Intervall und die temporaeren Grenzen
    private Interval interval;
    private Double aVariant;
    private Double bVariant;

    private Double slope;
    private Double yInterception;


    public TheilSenEstimator(LinkedList<Line> setOfLines, LinkedList<Point> setOfIntersections, Presenter presenter) {
        this.presenter = presenter;
        this.setOfLines = new ArrayList<>(setOfLines);
        this.setOfIntersections = new ArrayList<>(setOfIntersections);
        this.intervalIntersections = new ArrayList<>(setOfIntersections);

        this.n = setOfLines.size();
        this.sampledIntersections = new ArrayList<>();
        this.yCoordinates = new ArrayList<>();
        this.xCoordinates = new ArrayList<>();
        this.N = BinomialCoeffizient.run(n, 2);
        this.k = Integer.valueOf((int) (N * 0.5));

        //Koordinaten werden gespeichert damit am ende
        //der y Achsenabschnitt berechnet werden kann
        for (Point l : setOfIntersections) {
            yCoordinates.add(l.getY());
            xCoordinates.add(l.getX());
        }
    }

    public TheilSenEstimator(LinkedList<Line> setOfLines, LinkedList<Point> setOfIntersections) {
        this(setOfLines,setOfIntersections,null);
    }

    /**
     * Randomisierter Algorithmus zur Berechnung des Theil-Sen Schätzers.
     * Algorithmus stammt aus dem Paper:
     * "Jiri Matousek, Randomized optimal algorithm for slope selection,
     * Information Processing Letters 39 (1991) 183-187
     */
    public void run() {
        //damit eine initiale Ordnung herscht
        //Collections.sort(intervalIntersections);

        interval = new Interval(NEGATIV_INF, POSITIV_INF);
        while (true) {
            if (this.N <= n) {
                break;
            } else {
                r = Double.valueOf(n);
                //Anzahl der Schnittpunkte im Intervall [-Inf, a)
                int numberOfIntersections = getIntervalSize(NEGATIV_INF, interval.getLower(), setOfIntersections);

                //Randomized Interpolating Search
                j = (r / N) * (double) (k - numberOfIntersections);
                jA = (int) Math.max(1, Math.floor(j - (1.5 * Math.sqrt(r))));
                jB = (int) Math.min(r, Math.floor(j + (1.5 * Math.sqrt(r))));


                /* Suche nach einem passenderen und kleineren Intervall
                   Schleife terminiert wenn die das k-te Elemnet zwischen aVariant und bVariant liegt und
                   das Intrvall weniger als 11*N / sqrt(r) Elemente besitzt */
                do {
                    //zufällige Stichprobe
                    sampledIntersections = randomSampleOfIntersections(intervalIntersections, r);
                    //TODO: hier vlt. Randomized Select?!
                    /*Collections.sort(sampledIntersections);
                    aVariant = sampledIntersections.get(jA);
                    bVariant = sampledIntersections.get(jB);*/
                    aVariant = FastElementSelector.randomizedSelect(sampledIntersections, jA);
                    bVariant = FastElementSelector.randomizedSelect(sampledIntersections, jB);
                } while (!checkCondition());

                interval.setLower(aVariant);
                interval.setUpper(bVariant);
                N = Double.valueOf(getIntervalSize(interval.getLower(), interval.getUpper()));
                intervalIntersections = getIntervalElements(interval.getLower(), interval.getUpper());
            }
        }
    }


    /**
     * Diese Funktion überprüft ob die Bedingung für das Interval erfüllt ist. Dabei muss der k-te
     * Schnittpunkt in diesem Interval enthalten sein. des weiteren soll die Anzahl der Schnittpunkte
     * im Interval kleiner oder gleich dem Term: (11*N)/sqrt(r) sein.
     *
     * @return Boolscher Wert ob die Bedingung erfüllt ist
     */
    private Boolean checkCondition() {
        Double kthElement = FastElementSelector.randomizedSelect(xCoordinates, k);

        Boolean cond1 = (kthElement >= aVariant) && (
                kthElement < bVariant);
        Boolean cond2 = (getIntervalSize(aVariant, bVariant, intervalIntersections) <= ((11 * N) / Math.sqrt(r)));
        return cond1 && cond2;
    }


    /**
     * Diese Funktion gibt eine <code>r</code> Elementige Stichprobe aus der überegebenene Menge an
     * Schnittpunkten. Diese Stichprobe soll zufällig sein. Es können aus gleiche Werte in der Rückgabe
     * vertreten sein.
     *
     * @param set Menge an Schnittpunkten
     * @param r   Stichprobengröße
     * @return Stichprobe
     */
    public ArrayList<Double> randomSampleOfIntersections(ArrayList<Point> set, Double r) {
        ArrayList<Double> sampledLines = new ArrayList<>();
        Random random = new Random(System.currentTimeMillis());

        while (sampledLines.size() < r) {
            Double x = set.get(random.nextInt(set.size())).getX();
            sampledLines.add(x);

        }

        return sampledLines;

    }

    /**
     * Berechne wieviele von den Schnittpunkten in dem Interval zwischen <code>a</code> und <code>b</code>
     * enthalten sind.
     *
     * @param a untere Grenze des Intervals
     * @param b obere Grenze des Intrvals
     * @return Anzahl der Schnittpunkte im Interval [a,b)
     */
    public int getIntervalSize(double a, double b, ArrayList<Point> set) {
        int counter = 0;
        for (Point x : set) {
            if (x.getX() >= a && x.getX() < b) {
                counter++;
            }
        }
        return counter;
    }

    /**
     * Berechne wieviele von den Schnittpunkten in dem Interval zwischen <code>a</code> und <code>b</code>
     * enthalten sind.
     *
     * @param a untere Grenze des Intervals
     * @param b obere Grenze des Intrvals
     * @return Anzahl der Schnittpunkte im Interval [a,b)
     */
    public int getIntervalSize(double a, double b) {
        int counter = 0;
        for (int i=0;i<intervalIntersections.size();i++) {
            Point x = intervalIntersections.get(i);
            if (x.getX() >= a && x.getX() < b) {
                counter++;
            }
        }
        return counter;
    }

    /**
     * Berechne wieviele von den Schnittpunkten in dem Interval zwischen <code>a</code> und <code>b</code>
     * enthalten sind. Zusätzlich werden diese Schnittpunkte in einer Liste festgehalten und diese werden
     * zurückgeliefert.
     *
     * @param a untere Grenze des Intervals
     * @param b obere Grenze des Intrvals
     * @return Liste der Schnittpunkte die im Interval [a,b) vertreten sind
     */
    public ArrayList<Point> getIntervalElements(double a, double b) {
        ArrayList<Point> list = new ArrayList<>();
        for (int i=0;i<intervalIntersections.size();i++) {
            Point x = intervalIntersections.get(i);
            if (x.getX() >= a && x.getX() < b) {
                list.add(x);
            }
        }

        return list;

    }

    @Override
    public void getResult() {
        if (presenter != null) {
            setChanged();

            double m, x;
            double b, y;

            Collections.sort(xCoordinates);
            Collections.sort(yCoordinates);
            int n = xCoordinates.size();
            if (n % 2 == 0) {
                x = 0.5 * (xCoordinates.get((n / 2) - 1) + xCoordinates.get((n / 2)));
                y = 0.5 * (yCoordinates.get((n / 2) - 1) + yCoordinates.get((n / 2)));
            } else {
                x = xCoordinates.get(((n + 1) / 2) - 1);
                y = yCoordinates.get(((n + 1) / 2) - 1);
            }

            ArrayList<Point> resultSt = getIntervalElements(interval.getLower(), interval.getUpper());
            int size = resultSt.size();
            if (size % 2 == 0) {
                m = 0.5 * (resultSt.get((size / 2) - 1).getX() + resultSt.get((size / 2)).getX());
            } else {
                m = resultSt.get(((size + 1) / 2) - 1).getX();
            }

            b = (x * m) - y;

            slope = m;
            yInterception = b;

            String[] result = new String[]{"ts", m + "", b + ""};
            notifyObservers(result);
        }
    }


    public Double getSlope() {
        return slope;
    }

    public Double getyInterception() {
        return yInterception;
    }
}
